# File encoding: UTF-8

from more_itertools import peekable

__all__ = ["LogicError", "ItemError", "SIMPLE_MAP", "Item", "Node", "streem"]

class LogicError(RuntimeError):
    pass

class ItemError(RuntimeError):
    def __init__(self, what, item_value, item_level):
        self.item_value = item_value
        self.item_level = item_level
        super(ItemError, self).__init__(what, item_value, item_level)

SIMPLE_MAP = object()

_REDUCE_DEFAULT = list

class Item:
    __slots__ = ["value", "level", "level_rel", "next_level", "next_level_rel"]

    def __init__(self, value=None, level=None, level_rel=None, next_level=None, next_level_rel=None):
        self.value = value

        if level is not None and level_rel is not None:
            raise TypeError("level and level_rel can't both be present")

        if next_level is not None and next_level_rel is not None:
            raise TypeError("next_level and next_level_rel can't both be present")

        self.level = level
        self.level_rel = level_rel

        self.next_level = next_level
        self.next_level_rel = next_level_rel

    def get(self, current_level):
        item_level      = self.level      if self.level      is not None else current_level + (self.level_rel or 0)
        next_item_level = self.next_level if self.next_level is not None else item_level    + (self.next_level_rel or 0)
        return self.value, item_level, next_item_level

    def __repr__(self):
        return "Item({!r}, {})".format(self.value, ", ".join(("{}={:+d}" if k.endswith("R") else "{}={}").format(k, v) for k, v in [
            ("L",   self.level),
            ("LR",  self.level_rel),
            ("NL",  self.next_level),
            ("NLR", self.next_level_rel),
        ] if v is not None))

class Node:
    __slots__ = ["level", "value", "children"]

    def __init__(self, level, value, children):
        self.level = level
        self.value = value
        self.children = list(children) if isinstance(children, NodeIterator) else children

    def __repr__(self):
        return "Node({!r}, {!r}, {!r})".format(self.level, self.value, self.children)

    def as_text(self, indent=".", _nest_level=0):
        return "{}[{}] {}".format(str(indent) * _nest_level, self.level, self.value) + "".join("\n" + c.as_text(indent, _nest_level + 1) for c in self.children)

def with_levels(items, starting_level, mandatory_levels, mandatory_levels_all):
    level = starting_level
    last_level = starting_level - 1

    for item in items:
        item_value, item_level, item_next_level = item.get(level)

        if mandatory_levels_all:
            for l in range(last_level + 1, item_level):
                yield None, l
        else:
            for l in mandatory_levels: # must be sorted
                if last_level < l < item_level:
                    yield None, l

        if item_value is not None:
            yield item_value, item_level
            last_level = item_level

        level = item_next_level

class SourceData:
    __slots__ = ["mandatory_levels_max", "iter", "f_map", "f_reduce", "reduce_of_no_children", "ni_active"]

    def __init__(self, items, f_map, f_reduce, starting_level, mandatory_levels, mandatory_levels_all):
        self.mandatory_levels_max = max(mandatory_levels) if mandatory_levels else None
        self.iter = peekable(with_levels(items, starting_level=starting_level,
            mandatory_levels=(sorted(mandatory_levels) if not mandatory_levels_all else None),
            mandatory_levels_all=mandatory_levels_all,
        ))
        self.f_map = self.simple_struct_from_node if f_map is SIMPLE_MAP else (lambda x: x) if f_map is None else f_map
        self.f_reduce = f_reduce or _REDUCE_DEFAULT

        self.reduce_of_no_children = self.f_reduce([])

        self.ni_active = None

    def simple_struct_from_node(self, n):
        return (n.value, n.children) if n.children or (self.mandatory_levels_max is not None and n.level < self.mandatory_levels_max) else n.value

class NodeIterator:
    __slots__ = ["src", "level", "parent"]

    def __init__(self, src, parent=None):
        self.src = src
        self.level = self.src.iter.peek()[1]
        self.parent = parent

        self.src.ni_active = self

    def __iter__(self):
        return self

    def advance(self, consume_only=False):
        if self.src.ni_active is not self:
            raise LogicError("iterator inactive")

        item_value, item_level = self.src.iter.peek()
        assert item_level <= self.level
        if item_level < self.level:
            if self.parent is None or item_level > self.parent.level:
                raise ItemError("unexpected level {} (current {}, parent {})".format(item_level, self.level, self.parent.level if self.parent else None), item_value, item_level)
            raise StopIteration()

        next(self.src.iter)

        if self.src.iter.peek(default=(None, self.level))[1] <= self.level:
            if consume_only:
                return

            return self.src.f_map(Node(item_level, item_value, self.src.reduce_of_no_children))

        ni_children = NodeIterator(self.src, self)
        result = None if consume_only else self.src.f_map(Node(item_level, item_value, self.src.f_reduce(ni_children)))
        try:
            while True:
                ni_children.advance(consume_only=True)
        except StopIteration:
            self.src.ni_active = self
            return result

    def __next__(self):
        return self.advance()
    next = __next__

def streem(items, map=SIMPLE_MAP, reduce=None, starting_level=0, mandatory_levels=[], mandatory_levels_all=False):
    return (reduce or _REDUCE_DEFAULT)(NodeIterator(SourceData(items, map, reduce, starting_level, mandatory_levels, mandatory_levels_all)))




#!/usr/bin/python

import streem

import unittest
import re

s_test_data = """
===a                  | [0] None
====b                 | .[1] None
====c                 | ..[3] a
    +if(...)          | ...[4] b
     {+               | ...[4] c
      do_this         | ....[5] if(...)
      do_that         | ....[5] {
     -#}              | .....[6] do_this
     while(...)       | .....[6] do_that
     {+               | ....[5] while(...)
      {+              | ....[5] {
       some           | .....[6] {
       compound       | ......[7] some
       statement      | ......[7] compound
       4th            | ......[7] statement
       5th            | ......[7] 4th
      -#}             | .....[6] more
      more            | .....[6] statements
      statements      | .....[6] 4th
      4th             |
      5th             |
     -#}              |
     5th              |
=d                    | .[1] d
=========dd           | ..[3] None
===e                  | ...[9] dd
=======f              | ..[3] e
=======ff             | ...[7] f
=======fff            | ...[7] ff
=======ffff           | ...[7] fff
=======fffff          | ...[7] ffff
       +ignored       |
        +ignored      |
         ignored      |
       -ignored       |
=======ffffff         |
===h                  | ..[3] h
====i                 | ...[4] i
=================j++  | ....[17] j
--k================== | ....[17] k
-l                    | ....[17] l
m-                    | ....[17] m
+n (5th)              | ..[3] o
===o                  | ...[17] p
=================p    | ...[17] x
x+                    | ....[18] <hello>
 <hello>+             | .....[19] <world>
  <world>+            | ......[20] some
   some               | ......[20] text
   text               | ......[20] nodes
   nodes              | ......[20] 4th
   4th                | .....[19] <empty>
   5th                | ...[17] y
  -#</>               | ...[17] z
  <empty>+            |
  -#</>               |
 -#</>-               |
y                     |
z                     |
"""

re_line = re.compile(r"^([=]*)([+]*)([-]*)([#]?)([^#=+-]+?)([=]*)([+]*)([-]*)$")
def to_item(line):
    l, lrp, lrm, exclude, v, nl, nlrp, nlrm = re_line.match(line.strip()).groups()

    return streem.Item(value=(None if exclude else v.strip()),
        level=(len(l) or None),
        level_rel=(len(lrp) - len(lrm) or None),
        next_level=(len(nl) or None),
        next_level_rel=(len(nlrp) - len(nlrm) or None),
    )

class TestStreem(unittest.TestCase):

    def take(self, items, n):
        it = iter(items)
        try:
            for i in range(n):
                yield next(it)
        except StopIteration:
            self.assertRaises(StopIteration, next, it)
            raise StopIteration()

    def setUp(self):
        self.items = [to_item(line.partition("|")[0]) for line in s_test_data.strip().splitlines()]

    def test_result(self):
        nodes = streem.streem(
            self.items,
            map=None,
            reduce=lambda nodes: list(self.take(nodes, 4)),
            mandatory_levels=[0, 1, 3],
        )

        self.assertEqual(
            "\n".join(n.as_text() for n in nodes),
            "\n".join(x for x in (line.partition("|")[2].strip() for line in s_test_data.strip().splitlines()) if x).rstrip("\n"),
        )

    def test_level_enforcement(self):
        self.assertRaises(streem.ItemError, streem.streem,
            self.items,
            mandatory_levels=[0, 1],
        )

    def test_iterator_locking(self):
        all_inodes = []
        def reduce_storing_iterators(inodes):
            all_inodes.append(inodes)
            return inodes

        nodes = list(streem.streem(
            self.items,
            map=None,
            reduce=reduce_storing_iterators,
            mandatory_levels=[0, 1, 3],
        ))

        self.assertEqual(all_inodes[0], [])
        self.assertRaises(StopIteration, next, all_inodes[1])
        for it in all_inodes[2:]:
            if isinstance(it, list):
                self.assertEqual(it, [])
            else:
                self.assertRaises(streem.LogicError, next, it)

if __name__ == "__main__":
    unittest.main()




from distutils.core import setup

setup(
    name = "streem",
    packages = ["streem"],
    version = "1.0",
    description = "Turning streams into trees",
    author = "Roman Odaisky",
    author_email = "roma@qwertty.com",
    url = "https://github.com/rdaysky/streem",
    download_url = "https://github.com/rdaysky/streem/tarball/1.0",
    keywords = ["stream", "tree"],
    install_requires = ["more_itertools"],
    classifiers = [
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
    ],
)
